#!/usr/bin/bash

# vim:ft=sh:fdm=marker:fmr={,}

# archlabs installer library script file
# this file is not meant to be run directly
# sourcing this file in a non bash shell is not advised

format() {
    infobox "$_FSTitle" "\nFormatting: $1\n\nCommand: ${FS_CMDS[$2]}\n" 0
    ${FS_CMDS[$2]} $1 >/dev/null 2>$ERR
    check_for_errors "${FS_CMDS[$2]} $1"
}

decr_count() {
    # remove a partition from the dialog list and decrement the number partitions left
    local p="$1"
    PARTS="$(sed "s~${p} [0-9]*[G-M]~~; s~${p} [0-9]*\.[0-9]*[G-M]~~" <<< "$PARTS")"
    (( COUNT > 0 )) && (( COUNT-- ))
    return 0
}

enable_swap() {
    [[ $1 == "$MNT/swapfile" ]] && { fallocate -l $SWAP_SIZE $1; chmod 600 $1; }
    mkswap $1 >/dev/null 2>&1
    swapon $1 >/dev/null 2>&1
    return 0
}

mount_partition() {
    local part="$1"
    local mount="${MNT}$2"
    local fs="$(lsblk -lno FSTYPE $part)"

    mkdir -p "$mount"

    if [[ ${FS_OPTS[$fs]} != "" && $part != "$BOOT_PART" ]] && select_mount_opts "$part" "$fs"; then
        mount -o $MNT_OPTS $part "$mount" 2>$ERR
        check_for_errors "mount -o $MNT_OPTS $part $mount"
    else
        mount $part "$mount" 2>$ERR
        check_for_errors "mount $part $mount"
    fi

    confirm_mount $part "$mount" || return 1
    check_cryptlvm "$part"
    return 0
}

confirm_mount() {
    local part="$1"
    local mount="$2"

    if [[ "$mount" == "$MNT" ]]; then
        local msg="Partition: $part\nMountpoint: /  (root)"
    else
        local msg="Partition: $part\nMountpoint: ${mount#$MNT}"
    fi

    # partition failed to mount properly
    if ! grep -q "$mount" <<< "$(mount)"; then
        infobox "$_MntTitle" "$_MntFail\n$msg\n" 1
        return 1
    else
        # mount was successful
        infobox "$_MntTitle" "$_MntSucc\n$msg\n" 1
        decr_count "$part"
    fi

    return 0
}

find_partitions() {
    local str="$1"
    local err='NONE'

    # string of partitions as /TYPE/PART SIZE
    if [[ $IGNORE_DEV != "" ]]; then
        PARTS="$(lsblk -lno TYPE,NAME,SIZE | awk "/$str/"' && !'"/$IGNORE_DEV/"' {sub(/^part/, "/dev/"); sub(/^lvm|^crypt/, "/dev/mapper/"); print $1$2 " " $3}')"
    else
        PARTS="$(lsblk -lno TYPE,NAME,SIZE | awk "/$str/"' {sub(/^part/, "/dev/"); sub(/^lvm|^crypt/, "/dev/mapper/"); print $1$2 " " $3}')"
    fi

    # number of partitions total
    COUNT=$(wc -l <<< "$PARTS")

    # ensure we have enough partitions for the system and action type
    case $str in
        'part|lvm|crypt') [[ $COUNT -eq 0 || ($SYS == 'UEFI' && $COUNT -lt 2) ]] && err="$_PartErrBody" ;;
        'part|crypt') (( COUNT == 0 )) && err="$_LvmPartErrBody" ;;
        'part|lvm') (( COUNT < 2 )) && err="$_LuksPartErrBody" ;;
    esac

    # if there aren't enough partitions show the error message
    if [[ $err != 'NONE' ]]; then
        msgbox "$_ErrTitle" "$err"
        return 1
    fi
    return 0
}

check_cryptlvm() {
    local part="$1"
    local devs="$(lsblk -lno NAME,FSTYPE,TYPE)"

    # Identify if $part is "crypt" (LUKS on LVM, or LUKS alone)
    if [[ $(lsblk -lno TYPE "$part") =~ 'crypt' ]]; then
        LUKS=1
        LUKS_NAME="${part#/dev/mapper/}"

        for dev in $(awk '/lvm/ && /crypto_LUKS/ {print "/dev/mapper/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_DEV="$LUKS_DEV cryptdevice=$dev:$LUKS_NAME"
                LVM=1
                break
            fi
        done
        for dev in $(awk '/part/ && /crypto_LUKS/ {print "/dev/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_UUID="$(lsblk -lno UUID,TYPE,FSTYPE "$dev" | awk '/part/ && /crypto_LUKS/ {print $1}')"
                LUKS_DEV="$LUKS_DEV cryptdevice=UUID=$LUKS_UUID:$LUKS_NAME"
                break
            fi
        done

    elif [[ $(lsblk -lno TYPE "$part") =~ 'lvm' ]]; then
        LVM=1
        VOLUME_NAME="${part#/dev/mapper/}"

        for dev in $(awk '/crypt/ && /lvm2_member/ {print "/dev/mapper/"$1}' <<< "$devs" | uniq); do
            if grep -q "$VOLUME_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_NAME="$(sed 's~/dev/mapper/~~g' <<< "$dev")"
                break
            fi
        done
        for dev in $(awk '/part/ && /crypto_LUKS/ {print "/dev/"$1}' <<< "$devs" | uniq); do
            if grep -q "$LUKS_NAME" <<< "$(lsblk -lno NAME "$dev")"; then
                LUKS_UUID="$(lsblk -lno UUID,TYPE,FSTYPE "$dev" | awk '/part/ && /crypto_LUKS/ {print $1}')"
                LUKS_DEV="$LUKS_DEV cryptdevice=UUID=$LUKS_UUID:$LUKS_NAME"
                LUKS=1
                break
            fi
        done
    fi
}

auto_partition() {
    local device="$1"

    # confirm or bail
    yesno "$_PrepParts" "$_PartBody1 $device $_PartBody2" || return 0
    infobox "$_PrepParts" "\nAuto partitioning device: $device\n" 0

    swapoff -a  # make sure swap is disabled in case the device was used for swap

    local dev_info="$(parted -s $device print)"

    # walk the partitions on the device in reverse order and delete them
    for i in $(awk '/^ [1-9][0-9]?/ {print $1}' <<< "$dev_info" | sort -r); do
        parted -s $device rm $i >/dev/null 2>&1
    done

    # make sure we have the correct device table for the system type, gpt or msdos
    local newtable="gpt"
    local table="$(awk '/Table:/ {print $3}' <<< "$dev_info")"
    [[ $SYS == BIOS ]] && newtable="msdos"

    # if the current device table isn't correct run mklabel
    if [[ $table != "$newtable" ]]; then
        parted -s $device mklabel $newtable >/dev/null 2>&1
    fi

    # when device contains the string 'nvme' then add 'p' before the part number
    local nvme=""
    [[ $device =~ nvme ]] && nvme="p"

    local part_num=1
    BOOT_PART="$device${nvme}$part_num"  # set the boot partition label to the first partition

    if [[ $SYS == "BIOS" ]]; then
        parted -s $device mkpart primary ext4 1MiB 513MiB >/dev/null 2>&1
        mkfs.ext4 -q $BOOT_PART >/dev/null 2>&1
    else
        parted -s $device mkpart ESP fat32 1MiB 513MiB >/dev/null 2>&1
        mkfs.vfat -F32 $BOOT_PART >/dev/null 2>&1
    fi

    (( part_num++ ))                        # increment the partition number
    BOOT_DEVICE="$device"                   # only grub on BIOS systems uses this
    ROOT_PART="${device}${nvme}$part_num"   # set root partition label to the second partition

    parted -s $device mkpart primary ext4 514MiB 100% >/dev/null 2>&1
    mkfs.ext4 -q $ROOT_PART >/dev/null 2>&1

    tput civis
    sleep 0.5  # slow the process down, otherwise lsblk output can be wrong for some things

    echo -e "\nAuto partitioning complete.\n" > /tmp/.devlist
    lsblk $device -o NAME,MODEL,TYPE,FSTYPE,SIZE >> /tmp/.devlist
    dialog --cr-wrap --backtitle "$BT" --title " $_PrepParts " --textbox /tmp/.devlist 0 0
    return 0
}

setup_boot_device() {
    # set BOOT_DEVICE for syslinux on UEFI and grub on BIOS
    if [[ $BOOT_PART =~ nvme ]]; then
        BOOT_DEVICE="${BOOT_PART%p[1-9]}"
    else
        BOOT_DEVICE="${BOOT_PART%[1-9]}"
    fi

    BOOT_PART_NUM="${BOOT_PART: -1}"

    # setup the needed partition flags for boot on both system types
    parted -s $BOOT_DEVICE set $BOOT_PART_NUM boot on >/dev/null 2>&1
    if [[ $SYS == 'UEFI' ]]; then
        parted -s $BOOT_DEVICE set $BOOT_PART_NUM esp on >/dev/null 2>&1
    fi

    return 0
}

show_devices() {
    tput civis
    if [[ $IGNORE_DEV != "" ]]; then
        lsblk -o NAME,MODEL,TYPE,FSTYPE,SIZE,MOUNTPOINT |
            awk "!/$IGNORE_DEV/"' && /disk|part|lvm|crypt|NAME/ {print $0}' > /tmp/.devlist
    else
        lsblk -o NAME,MODEL,TYPE,FSTYPE,SIZE,MOUNTPOINT |
            awk '/disk|part|lvm|crypt|NAME/ {print $0}' > /tmp/.devlist
    fi
    dialog --cr-wrap --backtitle "$BT" --title " $_PrepShowDev " --textbox /tmp/.devlist 0 0
}

select_device() {
    local msg
    [[ $1 == 'boot' ]] && msg="$_DevSelTitle for bootloader\n" || unmount_partitions

    if (( DEV_COUNT == 1 )); then
        DEVICE="$(awk '{print $1}' <<< "$SYS_DEVS")"
        msg="\nOnly one device available$([[ $1 == 'boot' ]] && echo -n " for grub bootloader"):"
        infobox "$_DevSelTitle" "$msg $DEVICE\n" 1
    elif (( DEV_COUNT > 1 )); then
        tput civis
        DEVICE="$(dialog --cr-wrap --stdout --backtitle "$BT" --title " $_DevSelTitle " \
            --menu "${msg}$_DevSelBody" 0 0 0 $SYS_DEVS)"
        [[ $? != 0 || $DEVICE == "" ]] && return 1
    else
        msg="\nNo available devices for installation to use$([[ $1 == 'boot' ]] && echo -n " for grub bootloader")."
        msgbox "$_ErrTitle" "$msg\n$_Exit"
        die 1
    fi

    # if the device selected was for grub bootloader, set the BOOT_DEVICE
    # this is needed because grub uses the base device for BIOS, not the partition
    [[ $1 == 'boot' ]] && BOOT_DEVICE="$DEVICE"

    return 0
}

edit_partitions() {
    local device="$1"
    if [[ $device == "" ]]; then
        select_device 'root' || return 1
        device="$DEVICE"
    fi

    tput civis
    local choice
    choice="$(dialog --cr-wrap --stdout --backtitle "$BT" \
        --title " $_PartTitle " --menu "$_PartBody" 0 0 0 "$_PartAuto" "-" \
        $( ([[ $DISPLAY ]] && hash gparted >/dev/null 2>&1) && echo -n "gparted -") "cfdisk" "-"\
        "parted" "-" "$_PartWipe" "-")"
    [[ $? != 0 || $choice == "" ]] && return 1

    clear
    tput cnorm

    if [[ $choice != "$_PartWipe" && $choice != "$_PartAuto" ]]; then
        $choice $device
    elif [[ $choice == "$_PartWipe" ]]; then
        yesno "$_PartWipe" "$_PartBody1 $device $_PartWipeBody2" && { clear; wipe -Ifrev $device; }
        edit_partitions $device
    else
        # if auto_partition fails we need to re-initialize the variables, just to be sure
        auto_partition $device || { initialize_variables; return 1; }
    fi
}
